#!/usr/bin/python
# Copyright (C) 2009 Sugar Labs
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.

"""Activity information microformat parser.

Activity information is embedded in HTML/XHTML/XML pages using a
Resource Description Framework (RDF) http://www.w3.org/RDF/ .
 
An example::

<?xml version="1.0" encoding="UTF-8"?>
<RDF:RDF xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#" xmlns:em="http://www.mozilla.org/2004/em-rdf#"><RDF:Description about="urn:mozilla:extension:bounce">
    <em:updates>
        <RDF:Seq>
            <RDF:li resource="urn:mozilla:extension:bounce:7"/>
        </RDF:Seq>
    </em:updates>
</RDF:Description>

<RDF:Description about="urn:mozilla:extension:bounce:7">
    <em:version>7</em:version>
    <em:targetApplication>
        <RDF:Description>
            <em:id>{3ca105e0-2280-4897-99a0-c277d1b733d2}</em:id>
            <em:minVersion>0.82</em:minVersion>
            <em:maxVersion>0.84</em:maxVersion>
            <em:updateLink>http://activities.sugarlabs.org/downloads/file/25986/bounce-7.xo</em:updateLink>
            
            <em:updateHash>sha256:816a7c43b4f1ea4769c61c03fea24842ec5fa566b7d41626ffc52ec37b37b6c5</em:updateHash>
        </RDF:Description>
    </em:targetApplication>
</RDF:Description></RDF:RDF>

"""
from  xml.etree.ElementTree import ElementTree, XML

_DEBUG_PARSER = False
"""Check that the XML input is well-formed.
Otherwise, silently "do our best" with what we're given."""

class ASLOParser():
    """HTML parser to pull out data expressed in our microformat."""
    
    def __init__(self, xml_data):
        self.elem = XML(xml_data)

    def parse(self):
        try:
            self.version = self.elem.find(".//{http://www.mozilla.org/2004/em-rdf#}version").text
            self.link = self.elem.find(".//{http://www.mozilla.org/2004/em-rdf#}updateLink").text
        except:
            self.version = 0
            self.link = None

    def dump():
       iter = self.elem.getiterator()
       #Iterate
       for element in iter:
           #First the element tag name
            print "Element:", element.tag
            #Next the attributes (available on the instance itself using
            #the Python dictionary protocol
            if element.keys():
                print "\tAttributes:"
                for name, value in element.items():
                    print "\t\tName: '%s', Value: '%s'"%(name, value)
            #Next the child elements and text
            print "\tChildren:"
            #Text that precedes all child elements (may be None)
            if element.text:
                text = element.text
                text = len(text) > 40 and text[:40] + "..." or text
                print "\t\tText:", repr(text)
            if element.getchildren():
                #Can also use: "for child in element.getchildren():"
                for child in element:
                    #Child element tag name
                    print "\t\tElement", child.tag
                    #The "tail" on each child element consists of the text
                    #that comes after it in the parent element content, but
                    #before its next sibling.
                    if child.tail:
                        text = child.tail
                        text = len(text) > 40 and text[:40] + "..." or text
                        print "\t\tText:", repr(text)  

def parse_aslo(xml_data):
    """Parse the activity information embedded in the given string
    containing XML data.  Returns a list containing the activity version and url.
    """
    ap = ASLOParser(xml_data)
    ap.parse()
    return ap.version, ap.link

def parse_url(url, **urlopen_args):
    """Parse the activity information at the given URL. Returns the same
    information as `parse_xml` does, and raises the same exceptions.
    The `urlopen_args` can be any keyword arguments accepted by
    `bitfrost.util.urlrange.urlopen`."""
    import bitfrost.util.urlrange as urlrange
    with urlrange.urlopen(url, **urlopen_args) as f:
        return parse_aslo(f.read())

#########################################################################
# Self-test code.
def _main():
    """Self-test."""
    print parse_url('http://activities.sugarlabs.org/services/update.php?id=bounce')

if __name__ == '__main__': _main ()
