#!/usr/bin/python
# Copyright (C) 2008 One Laptop Per Child Association, Inc.
# Licensed under the terms of the GNU GPL v2 or later; see COPYING for details.
# Written by C. Scott Ananian <cscott@laptop.org>
"""Activity updater: backing model.

This module implements the non-GUI portions of the activity updater,
including in particular the master list of groups, activities, whether
updates are needed, and the URL at which to find the updated activity.

Because `UpdateList` inherits from `gtk.ListStore` so that it plays
nicely with a GUI, this module requires `gtk`.  Those of you without
DISPLAY set will have to put up with a GtkWarning that the display
couldn't be opened.  Sorry.
"""
from __future__ import with_statement
from __future__ import division

# for testing
_DEBUG_MAKE_ALL_OLD = True

# default timeout for HTTP connections, in seconds
HTTP_TIMEOUT=30

import gtk
import gobject

import locale
import os
import os.path
import socket
import sys
import traceback
import zipfile
from HTMLParser import HTMLParseError
from urllib2 import HTTPError

import gettext
_ = lambda msg: gettext.dgettext('sugar-update-control', msg)

import bitfrost.update.actinfo as actinfo
import bitfrost.update.actutils as actutils
import ASLOParser as ASLOParser
import bitfrost.util.urlrange as urlrange

# weak dependency on inhibit_suspend from olpc-update package
try:
    from bitfrost.update import inhibit_suspend
except ImportError:
    # use a no-op substitude for the inhibit_suspend decorator.
    inhibit_suspend = lambda f: f

##########################################################################
# Fundamental data object: the real work gets done here.

_column_name_map = dict(globals())
ACTIVITY_ID, ACTIVITY_BUNDLE, ACTIVITY_ICON, \
             UPDATE_URL, UPDATE_VERSION, UPDATE_SIZE, \
             UPDATE_EXISTS, UPDATE_SELECTED, \
             DESCRIPTION_BIG, DESCRIPTION_SMALL, IS_HEADER = \
             xrange(11)
"""List of columns in the `UpdateList`."""
_column_name_map = dict((k,v) for k,v in globals().items()
                        if k not in _column_name_map and k!='_column_name_map')
"""Mapping from column names to indices."""

class UpdateList(gtk.ListStore):
    """Model which provides backing storage for the activity list treeview."""
    __gproperties__ = {
        'is_valid': (gobject.TYPE_BOOLEAN, 'is valid',
                     'true iff the UpdateList has been properly refreshed',
                     False, gobject.PARAM_READABLE),
        'saw_network_failure': (gobject.TYPE_BOOLEAN, 'saw network failure',
                                'true iff at least one network IO error '+
                                'occurred when the UpdateList was last '+
                                'refreshed',
                                False, gobject.PARAM_READABLE),
        'saw_network_success': (gobject.TYPE_BOOLEAN, 'saw network success',
                                'true iff at least one network operation '+
                                'completed successfully when the UpdateList '+
                                'was last refreshed',
                                False, gobject.PARAM_READABLE),
    }

    def __init__(self, skip_icons=False):
        gtk.ListStore.__init__(self,
                               # column types
                               str, object, gtk.gdk.Pixbuf,
                               str, long, long,
                               bool, bool, str, str, bool)

        self._parse_cache = {}
        self._skip_icons = skip_icons
        self._is_valid = False
        self._network_failures = []
        self._saw_network_success = False

    def __del__(self):
        """Free up any memory held by the cache in urlrange."""
        urlrange.urlcleanup()

    def _append(self, at=None, **kwargs):
        """Utility function to make it easier to add rows and get paths."""
        global _column_name_map
        # defaults for each column
        row = [None, None, None,
               None, 0, 0,
               True, True, None, None, False]
        # set entries in the row based on kwargs
        for k,v in kwargs.items():
            row[_column_name_map[k]] = v
        if at is not None:
            it = self.insert(at, row)
        else:
            it = self.append(row)
        return self.get_path(it)

    def toggle_select(self, path):
        """Toggle whether the given update will be installed."""
        row = self[path]
        row[UPDATE_SELECTED] = not row[UPDATE_SELECTED]

    # don't touch the UI in refresh, it needs to be thread-safe.
    # (model will be disconnected from the view before invoking refresh
    #  in another thread)
    def refresh_list(self, progress_callback=lambda n, extra: None,
                clear_cache=True):
        """Perform network operations to find available updates.

        The `progress_callback` is invoked with numbers between 0 and 1
        or `None` as the network queries complete.  The last callback will be
        `progress_callback(1)`.  Passing `None` to `progress_callback`
        requests "pulse mode" from the progress bar.
        """
        global _parse_cache
        if clear_cache:
            _parse_cache = {} # clear ASLOParser parse cache
            urlrange.urlcleanup() # clean url cache
        self._cancel = False
        self._invalidate()
        # don't notify for the following; we'll notify at the end when we
        # know what the new values ought to be.
        self._saw_network_success = False
        self._network_failures = []
        # bookkeeping
        progress_callback(None, None)
        self.clear()
        # find all activities already installed.
        progress_callback(None, _('Looking for local activities and content...')) # pulse
        activities = actinfo.get_activities() + actinfo.get_libraries()

        # now we've got enough information to allow us to compute a
        # reasonable completion percentage.
        steps_total = [ len(activities) + 3 ]
        steps_count = [ 0 ] # box this to allow update from mkprog.
        def mkprog(msg=None):
            """Helper function to do progress update."""
            steps_count[0] += 1
            progress_callback(steps_count[0]/steps_total[0], msg)

        row_map = {}

        # add in information from local activities.
        self._append(IS_HEADER=True, DESCRIPTION_BIG=_('Local activities'))
        for act in activities:
            act_id = act.get_bundle_id()
            if act_id not in row_map:
                row_map[act_id] = self._append(ACTIVITY_ID=act_id,
                                               UPDATE_EXISTS=False)
            else:
                steps_total[0] -= 1 # correct double-counting.
            # update icon, and bundle
            row = self[row_map[act_id]]
            row[ACTIVITY_BUNDLE] = act
            row[DESCRIPTION_BIG] = act.get_name()
            if not self._skip_icons:
                try:
                    row[ACTIVITY_ICON] = _svg2pixbuf(act.get_icon_data())
                except IOError:
                    # dlo trac #8149: don't kill updater if existing icon
                    # bundle is malformed.
                    pass
        
        for row in self:
            if self._cancel: break # bail!
            if row[IS_HEADER]: continue # skip
            mkprog(_('Checking %s...') % row[DESCRIPTION_BIG])
            try:
                self.refresh_row(row)
            except:
                row[UPDATE_EXISTS] = False # something wrong, can't update
                if row[UPDATE_URL] is not None:
                    self._network_failures.append(row[UPDATE_URL])
                # log the problem for later debugging.
                print "Failure updating", row[DESCRIPTION_BIG], \
                      row[DESCRIPTION_SMALL], row[UPDATE_URL]
                traceback.print_exc()
        mkprog('Sorting...') # all done
        # hide headers if all children are hidden
        sawone, last_header = False, None
        for row in self:
            if row[IS_HEADER]:
                if last_header is not None:
                    last_header[UPDATE_EXISTS] = sawone
                sawone, last_header = False, row
            elif row[UPDATE_EXISTS]:
                sawone = True
        if last_header is not None:
            last_header[UPDATE_EXISTS] = sawone
        # finally, sort all rows.
        self._sort()
        mkprog() # all done
       
        self._is_valid = True
        self.notify('is-valid')
        self.notify('saw-network-failure')
        self.notify('saw-network-success')

    def refresh_row(self, row):
        """Look for updates to an existing activity."""
        act = row[ACTIVITY_BUNDLE]
        def net_good(url_): self._saw_network_success = True
        def net_bad(url): self._network_failures.append(url)
        oldver, newver, newurl, size = \
                self._retrieve_update_version(act, net_good, net_bad)
        # merge w/ info from the activity group
        # (activity group entries have UPDATE_EXISTS=True)
        if row[UPDATE_EXISTS] and \
           ((act.get_activity_version() if newver is None else newver) \
            < row[UPDATE_VERSION]):
            newver, newurl = row[UPDATE_VERSION], row[UPDATE_URL]
            size = urlrange.urlopen(row[UPDATE_URL], timeout=HTTP_TIMEOUT)\
                   .length()
        row[UPDATE_EXISTS] = (newver is not None)
        row[UPDATE_URL] = newurl
        row[UPDATE_SIZE] = size
        if newver is None:
            description = _('At version %s') % oldver
        else:
            description = \
                _('From version %(old)d to %(new)s (Size: %(size)s)') % \
                { 'old':oldver, 'new':newver, 'size':_humanize_size(size) }
            row[UPDATE_SELECTED] = True
        row[DESCRIPTION_SMALL] = description
    
    def _retrieve_update_version(self, actbun,
                             network_success_cb=(lambda url: None),
                             network_failure_cb=(lambda url: None)):
        """Return a tuple of current version, new version, url for new
        version, size of new version, given a current activity bundle.
        All the information about the new version is `None` if no newer
        update can be found.  The optional `network_success_cb` is invoked
        with a single `url` parameter whenever a network operation
        succeeds, and the optional `network_failure_cb` is invoked with a
        single `url` parameter if all fallback network operations for a
        given activity bundle fail."""
        
        oldv = 0 if _DEBUG_MAKE_ALL_OLD else actbun.get_activity_version()
       
        
        newv, newu = self._check_for_updates(actbun.get_bundle_id())
        #    if vlist is not None:
        #        network_success_cb(uu)
        #        if len(vlist) > 0:
        #            break # found some info
        #if vlist is None:
        #    # all url variants for the given actbun have failed.
        #    network_failure_cb(update_url)
        #    return oldv, None, None, 0
        
        if newu is None or newv==oldv: return oldv, None, None, 0 # no updates
        # find size of new version.
        try:
            size = urlrange.urlopen(newu, timeout=HTTP_TIMEOUT).length()
            return oldv, newv, newu, size
        except (HTTPError, IOError, socket.error):
            # hmm, i guess that url isn't valid after all.  bail.
            return oldv, None, None, 0 # there are no *actual* updates.

    def _check_for_updates(self, activity_id):
        """Downloads the given URL, parses it (caching the result), and returns
        a list of (version, url) pairs present for the given `activity_id`.
        Returns `None` if there was a problem downloading the URL.
        Returns a zero-length list if the given URL is unparsable or does not
        contain information for the desired activity_id."""
        
        #Skip caching for now
        #if url not in self._parse_cache:
        #    try:
        #        __, __, self._parse_cache[url] = \
        #            ASLOParser.parse_url(url, timeout=HTTP_TIMEOUT)
        #    except HTMLParseError:
        #        self._parse_cache[url] = {} # parse error
        #    except (IOError, socket.error):
        #        self._parse_cache[url] = None # network error
        #activity_map = self._parse_cache[url]
        #if activity_map is None:
        #    return None # error attempting to check.
        #if activity_id not in activity_map:
        #    return [] # no versions found.
        #return activity_map[activity_id]
        
        url = 'http://activities.sugarlabs.org/services/update.php?id=' + activity_id
        return ASLOParser.parse_url(url, timeout=HTTP_TIMEOUT)

    def _sort(self):
        """Sort rows by group number, then case-insensitively by description."""
        l = lambda s: s.decode('utf-8','ignore').lower() \
            if s is not None else None
        def sort_value(row_num):
            row = self[row_num]
            return (0 if row[IS_HEADER] else 1,
                    l(row[DESCRIPTION_BIG]), l(row[DESCRIPTION_SMALL]))
        row_nums = range(len(self))
        row_nums.sort(key=sort_value)
        self.reorder(row_nums)

    def cancel_refresh(self):
        """Asynchronously cancel a `refresh` operation."""
        self._cancel = True

    def cancel_download(self):
        """Asynchronously cancel a `download_selected_updates` operation."""
        self._cancel = True

    def _sum_rows(self, row_func):
        """Sum the values returned by row_func called on all non-header
        rows."""
        return sum(row_func(r) for r in self if not r[IS_HEADER])

    def updates_available(self):
        """Return the number of updates available.

        Updated by `refresh`."""
        return self._sum_rows(lambda r: 1 if r[UPDATE_EXISTS] else 0)

    def updates_selected(self):
        """Return the number of updates selected."""
        return self._sum_rows(lambda r: 1 if
                              r[UPDATE_EXISTS] and r[UPDATE_SELECTED] else 0)
    def updates_size(self):
        """Returns the size (in bytes) of the selected updates available.

        Updated by `refresh`."""
        return self._sum_rows(lambda r: r[UPDATE_SIZE] if
                              r[UPDATE_EXISTS] and r[UPDATE_SELECTED] else 0)
    def unselect_all(self):
        """Unselect all available updates."""
        for row in self:
            if not row[IS_HEADER]:
                row[UPDATE_SELECTED] = False

    def select_all(self):
        """Select all available updates."""
        for row in self:
            if not row[IS_HEADER]:
                row[UPDATE_SELECTED] = True

    def is_valid(self):
        """The UpdateList is invalidated before it is refreshed, and when
        the group information is modified without refreshing."""
        return self._is_valid

    def _invalidate(self):
        """Set the 'is-valid' property to False and notify listeners."""
        if self._is_valid:
            # don't notify if already invalid
            self._is_valid = False
            self.notify('is-valid')

    def saw_network_failure(self):
        """Returns true iff there was at least one network failure during the
        last refresh operation."""
        return len(self._network_failures) > 0

    def saw_network_success(self):
        """Returns true iff there was at least one successful network
        transaction during the last refresh operation."""
        return self._saw_network_success

    def do_get_property(self, prop):
        """Standard interface to access the 'is-valid' property."""
        if prop.name == 'is-valid': return self.is_valid()
        if prop.name == 'saw-network-failure': return self.saw_network_failure()
        if prop.name == 'saw-network-success': return self.saw_network_success()
        raise AttributeError, 'unknown property %s' % prop.name

    
    def download_selected_updates(self, progress_cb=(lambda n, row: None),
                                  dir=None):
        """Return a generator giving (row, local filename) pairs for
        each selected update.  Caller is responsible for unlinking
        these files when they are through.  The `progress_cb` gets a
        floating point number in [0, 1] indicating the amount of the
        download which is complete, and a reference to the row in this
        `UpdateList` describing the current download.  The files will
        be created in the directory specified by `dir`, if it is
        given, otherwise they will be created in the user's activity
        directory.  If the remote HTTP server does not provide information
        about the length of some of the downloads, the first parameter
        to the `progress_cb` will be `None` during those downloads.
        """
        if dir is None: dir=actinfo.USER_ACTIVITY_DIR
        if not os.path.isdir(dir): os.makedirs(dir)
        self._cancel = False
        sizes = [ 0, self.updates_size() ]
        for row in self:
            if self._cancel: return # bail.
            if row[IS_HEADER]: continue
            if not (row[UPDATE_EXISTS] and row[UPDATE_SELECTED]): continue
            def report(cursize, totalsize_):
                if totalsize_ is None:
                    progress_cb(None, row)
                else:
                    progress_cb((sizes[0]+cursize) / sizes[1], row)
            try:
                yield row, urlrange.urlretrieve(row[UPDATE_URL], dir=dir,
                                                reporthook=report,
                                                cancelhook=lambda:self._cancel,
                                                store_cache=False,
                                                timeout=HTTP_TIMEOUT)
            except urlrange.CancelException:
                return # bail.
            except (HTMLParseError, IOError, socket.error):
                yield row, None # network error
            # XXX: if rows eventually have 'indeterminate size', then we
            #      should update this progress computation, probably based
            #      on the size of the file we just yielded above!
            sizes[0] += row[UPDATE_SIZE]
            progress_cb(sizes[0] / sizes[1], row)

#########################################################################
# Utility Funtions

def _humanize_size(bytes):
    """
    Convert a given size in bytes to a nicer better readable unit
    """
    if bytes == 0:
        # TRANSLATORS: download size is 0
        return _("None")
    elif bytes < 1024:
        # TRANSLATORS: download size of very small updates
        return _("1 KB")
    elif bytes < 1024 * 1024:
        # TRANSLATORS: download size of small updates, e.g. "250 KB"
        return locale.format(_("%.0f KB"), bytes/1024)
    else:
        # TRANSLATORS: download size of updates, e.g. "2.3 MB"
        return locale.format(_("%.1f MB"), bytes / 1024 / 1024)

def _svg2pixbuf(icon_data):
    """Convert the given `icon_data` SVG string to a `gtk.gdk.Pixbuf`
    with maximum size 55x55."""
    import re, rsvg
    # substitute black/white for icon color entities.
    for entity, value in [('stroke_color','#808080'),
                          ('fill_color','#eee')]:
        xml = '<!ENTITY %s "%s">' % (entity, value) 
        icon_data = re.sub('<!ENTITY %s .*>' % entity, xml, icon_data) 
    h = rsvg.Handle(data=icon_data)
    if h.get_property('width') > 55 or h.get_property('height') > 55:
        # lame! scale it.
        print "WARNING: oversize icon (%dx%d), scaling." % \
              (h.get_property('width'), h.get_property('height'))
        del h
        pbl = gtk.gdk.PixbufLoader()
        pbl.set_size(55, 55)
        pbl.write(icon_data)
        pbl.close()
        return pbl.get_pixbuf()
    return h.get_pixbuf()

#########################################################################
# Control panel keys for command-line and GUI use.

# note that the *presence* of a key is defined by the presence of a
# get_<key> method, the *documentation* for a key is given by the
# doc string of the set_<key> method, but actually looking at
# values from the command-line is performed using print_<key>.

def _print_status(n, extra):
    """Returns a helper function which print a human-readable
    completion percentage."""
    sys.stdout.write(' '*79)
    sys.stdout.write('\r')
    if extra is None: extra=''
    if n is None:
        sys.stdout.write('%s\r' % extra[:77].ljust(77))
    else:
        sys.stdout.write('%s %5.1f%%\r' % (extra[:70].ljust(70), n*100))
    sys.stdout.flush()

# 'available_updates' key -----------------------------
def get_available_updates():
    # only here to get this key to show up for '-l'
    raise ValueError('This method not used by the GUI')

@inhibit_suspend
def print_available_updates():
    """Print the set of available updates in nice human-readable fashion."""
    ul = UpdateList(skip_icons=True)
    ul.refresh_list(_print_status)
    print
    def opt(x):
        if x is None or x == '': return ''
        return ': %s' % x
    for row in ul:
        if not row[UPDATE_EXISTS]: continue # skip
        if row[IS_HEADER]:
            print row[DESCRIPTION_BIG] + opt(row[DESCRIPTION_SMALL])
        else:
            print '*', row[DESCRIPTION_BIG] + opt(row[DESCRIPTION_SMALL])
    print
    print _('%(number)d updates available.  Size: %(size)s') % \
          { 'number': ul.updates_available(),
            'size': _humanize_size(ul.updates_size()) }

def set_available_updates():
    # NOTE slightly odd way to document this control panel key
    """Retrieve the list of available activity updates."""
    raise ValueError(_("Setting the list of updates is not permitted."))

# 'install_updates' key -----------------------------
def get_install_update():
    # this func is needed to make the 'install_update' key show up for '-l'
    raise ValueError(_("Only the 'set' operation for this key is defined."))

@inhibit_suspend
def set_install_update(which):
    """
    Install any available updates for the activity identified by the
    given (localized) name or bundle id, or all available updates if
    'all' is given.
    """
    import os
    ul = UpdateList(skip_icons=True)
    ul.refresh_list(_print_status)
    def too_many():
        raise ValueError(_('More than one match found for the given activity name or id.'))
    def no_updates():
        raise ValueError(_('The given activity is already up-to-date.'))
    if which != 'all':
        found = False
        for row in ul:
            row[UPDATE_SELECTED] = False
        for row in ul:
            if row[IS_HEADER]: continue
            if which==row[ACTIVITY_ID]:
                if found: too_many()
                found = True
                if not row[UPDATE_EXISTS]: no_updates()
                row[UPDATE_SELECTED] = True
        if not found:
            found_but_uptodate=False
            for row in ul:
                if row[IS_HEADER]: continue
                if which in row[DESCRIPTION_BIG]:
                    if not row[UPDATE_EXISTS]: # could be false match
                        found_but_uptodate=True
                        continue
                    if found: too_many()
                    found = True
                    row[UPDATE_SELECTED] = True
        if not found:
            if found_but_uptodate: no_updates()
            raise ValueError(_('No activity found with the given name or id.'))
        assert ul.updates_selected() == 1
    # okay, now we've selected only our desired updates.  Download and
    # install them!
    # we need to set up a glib event loop in order to connect to the
    # activity registry (sigh) in ActivityBundle.upgrade() below.
    from dbus.mainloop.glib import DBusGMainLoop
    DBusGMainLoop(set_as_default=True)
    # we'll fetch the main loop, but we don't actually have to run it.
    loop = gobject.MainLoop()
    from jarabe.model.bundleregistry import get_registry
    registry = get_registry()
    def reporthook(n, row):
        _print_status(n, _('Downloading %s...') % row[DESCRIPTION_BIG])
    for row, f in ul.download_selected_updates(reporthook):
        if f is None: continue # cancelled or network error
        try:
            _print_status(None, _('Examining %s...') % row[DESCRIPTION_BIG])
            b = actutils.BundleHelper(f)
            if b.is_installed(registry):
                _print_status(None, _('Upgrading %s...') % row[DESCRIPTION_BIG])
            else:
                _print_status(None, _('Installing %s...')% row[DESCRIPTION_BIG])
            b.install_or_upgrade(registry)
        except:
            print
            print _('Error installing %s.') % row[DESCRIPTION_BIG]
            import traceback
            traceback.print_exc() # complain!  but go on.
        if os.path.exists(f):
            os.unlink(f)
        else:
            print "Failed trying to clean up", f


#########################################################################
# Self-test code.
def _main():
    """Self-test."""
    print_available_updates()
    #set_install_update('all')
    #ul = UpdateList(skip_icons=True)
    #print ul._check_for_updates('bounce')

if __name__ == '__main__': _main ()
